/*
	kopen		Open the serial port for K-Team robots
	
	The calling syntax is:
		
		ref = kopen([port,bps,timeout])
	
	Arguments:
	
		port:		0 for the modem port, 1 for the printer port (default = 0)
		bps:		speed (typically 9600, 19200, 38400, ...)
		timeout:	timeout in seconds
			(the default values are [0, 9600, 1])
		ref:		a reference to the serial port which should be used with ksend
					and kclose (can be a matrix)

	Under Windows, compile by typing in Matlab's command line

		mex -output kopen KOpenMEX.c KhepComWin.c
	
	Copyright 1998, Yves Piguet
	All Righs Reserved
*/

//#include "CodeWarrior.MEX.prefix"

#include "mex.h"

#include "KhepCom.h"

/* Input Arguments */

#define	CONFIG_IN	prhs[0]

/* Output Arguments */

#define	PORTREF_OUT	plhs[0]

// Main Function

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	short port;
	long rate;
	double timeout;
	KPortRef portRef;
	double *ref;	// same, in the output argument
	int err;
	
	/* Checks for proper number of arguments */
	
	if (nrhs > 1 || nlhs > 1)
		mexErrMsgTxt("KOPEN syntax: ref=kopen(config) or ref=kopen");
	
	/* Gets the parameters */
	
	if (nrhs == 1)
	{
		if(!mxIsDouble(CONFIG_IN) || mxIsComplex(CONFIG_IN) ||
									mxGetM(CONFIG_IN) != 1 || mxGetN(CONFIG_IN) > 3) 
			mexErrMsgTxt("KOPEN arg. must be a row vector of up to three numbers");
		port = (short)mxGetPr(CONFIG_IN)[0];
		if (port != 0 && port != 1)
			mexErrMsgTxt("KOPEN port must be 0 (modem) or 1 (printer)");
		if (mxGetN(CONFIG_IN) >= 2)
			rate = (long)mxGetPr(CONFIG_IN)[1];
		else
			rate = 19200;
		if (mxGetN(CONFIG_IN) >= 3)
			timeout = mxGetPr(CONFIG_IN)[2];
		else
			timeout = 1;
	}
	else
	{
		port = 0;
		rate = 19200;
		timeout = 1;
	}
	
	err = KOpenSerial(port, rate,
					timeout < 0 ? 60000L : timeout > 60 ? 60000L : (long)(timeout * 1000),
					&portRef);
	
	switch (err)
	{
		case kKhepNoErr:
			PORTREF_OUT = mxCreateDoubleMatrix(1, 4, 0);
			ref = mxGetPr(PORTREF_OUT);
			ref[0] = portRef.priv[0];
			ref[1] = portRef.priv[1];
			ref[2] = portRef.priv[2];
			ref[3] = portRef.priv[3];
			break;
		case kKhepDeviceErr:
			mexErrMsgTxt("KOPEN device error");
			break;
		case kKhepComErr:
			mexErrMsgTxt("KOPEN read/write error");
			break;
		case kKhepComTimeoutErr:
			mexErrMsgTxt("KOPEN timeout error");
			break;
		case kKhepComBadPortErr:
			mexErrMsgTxt("KOPEN unsupported serial port");
			break;
		case kKhepComBadBaudRateErr:
			mexErrMsgTxt("KOPEN unsupported baud rate");
			break;
	}
}
