/*
 *	SQExtFunctionsKhep.c
 *
 *	Additional SysQuake functions for Khepera
 *
 *	Copyright 1999-2000, Yves Piguet and K-Team
 */

#include "LME.h"
#include "SQXFn.h"
#include "SQShLibXFn.h"
#include "SQLaunchURLXFn.h"
#include "SQKhep.h"
#include "KhepCom.h"

//#define kCurrentVersion 10029	// 1.00b9 (see SQXFn.h)
#define kCurrentVersion 11000	// 1.1 (see SQXFn.h)

static LMEInt gFirstFreeId;
static LMEWriteCB oldWriteCB;
static LMEReadCB oldReadCB;

#define maxc 16
static struct
{
	KPortRef ref;
	LMEBoolean textMode;	// replace crlf and lf with lf
	LMEBoolean wasCR;		// true if the last character read was CR
	LMEBoolean inUse;
} c[maxc];

static LMEErr info(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)	// test
{
	(void)nargin, nargout, user;
	
	return LME_PushString(lme, 31, (LMEConstText)"SysQuake customized for K-Team\n");
}

static void InitDevices(void)
{
	LMEInt i;
	
	for (i = 0; i < maxc; i++)
		c[i].inUse = 0;
}

static void ShutdownDevices(void)
{
	LMEInt i;
	
	for (i = 0; i < maxc; i++)
		if (c[i].inUse)
			KCloseSerial(&c[i].ref);
}

static LMEErr opendevice(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
	// fd = opendevice(devicename, options)
	// devicename (string): name of the device to open (such as kserialname(i))
	// options (string): comma-separated list of options:
	//		bn: bps (where n is a valid bps value)
	//		Tn: timeout in ms (0 = none)
	//		t:  text mode (cr and crlf are replaced with lf)
{
	LMEInt i, j, port, nameLength, nameLength2, optionsLength;
	LMEText name, options;
	char name2[kNameMaxLength];
	LMEInt bps, timeout;
	LMEBoolean textMode;
	LMEErr err;
	(void)nargout, user;
	
	// get the device name
	err = LME_ConvertStringInPlace(lme, nargin - 1, &name, &nameLength);
	if (err != kLMEErrOk)
		return err;
	
	// get the options, if any
	bps = 19200;
	timeout = 1000;
	textMode = 0;
	if (nargin > 1)
	{
		err = LME_ConvertStringInPlace(lme, 0, &options, &optionsLength);
		if (err != kLMEErrOk)
			return err;
		for (i = 0; i < optionsLength; )
		{
			switch (options[i])
			{
				case ',':
					i++;
					break;
				case 'b':	// b - baud rate
					for (i++, bps = 0; i < optionsLength
								&& options[i] >= '0' && options[i] <= '9'; i++)
						bps = 10 * bps + options[i] - '0';
					break;
				case 'T':	// T - timeout [ms]
					for (i++, timeout = 0; i < optionsLength
								&& options[i] >= '0' && options[i] <= '9'; i++)
						timeout = 10 * timeout + options[i] - '0';
					break;
				case 't':	// t - text mode (replace cr and crlf with lf)
					textMode = 1;
					i++;
					break;
				default:
					i++;
					break;
			}
		}
	}
	
	// find a free connection slot
	for (i = 0; i < maxc && c[i].inUse; i++)
		;
	if (i >= maxc)
		return kLMEErrTooManyFiles;
	
	// find the device number
	for (port = 0; ; port++)
	{
		if (!KGetSerialDeviceName(port, name2, &nameLength2))
			return kLMEErrFileNotFound;
		if (nameLength == nameLength2)
		{
			for (j = 0; j < nameLength && name[j] == name2[j]; j++)
				;
			if (j == nameLength)	// ok
				break;
		}
	}
	
	// open port
	switch (KOpenSerial(port, bps, timeout, &c[i].ref))
	{
		case kKhepNoErr:
			c[i].textMode = textMode;
			c[i].wasCR = 0;
			c[i].inUse = 1;
			return LME_PushReal(lme, gFirstFreeId + i);
			break;
		case kKhepDeviceErr:
			return kLMEErrBadAccess;
		case kKhepComErr:
			return kLMEErrBadAccess;
		case kKhepComTimeoutErr:
			return kLMEErrTimeout;
		case kKhepComBadPortErr:
			return kLMEErrFileNotFound;
		case kKhepComBadBaudRateErr:
			return kLMEErrBadArg;
	}
	
	return kLMEErrBadAccess;
}

static LMEErr closedevice(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
{
	LMEErr err;
	LMEFloat x;
	(void)nargin, nargout, user;
	
	err = LME_PopReal(lme, &x);
	if (err != kLMEErrOk)
		return err;
	
	if ((LMEInt)x < gFirstFreeId || (LMEInt)x >= gFirstFreeId + maxc
											|| !c[(LMEInt)x - gFirstFreeId].inUse)
		return kLMEErrBadFd;
	
	KCloseSerial(&c[(LMEInt)x - gFirstFreeId].ref);
	c[(LMEInt)x - gFirstFreeId].inUse = 0;
	return kLMEErrOk;
}

static LMEXFunctionStruct optXf[] =
{
//	{info, 0, 0, "info", 0, 0, 1, 1},	// test
	
#ifdef macintosh
	XFOpenShLib,
	XFCallShLib,
	XFOldCallShLib,
	XFCloseShLib,
	
	XFLaunchURL,
#endif
	
	{opendevice, 0, 0, "opendevice", 1, 2, 1, 1},
	{closedevice, 0, 0, "closedevice", 1, 1, 0, 0},
	
	XFKSerialName,
	XFKOpen,
	XFKSend,
	XFKFlush,
	XFKClose,
	
	LMEXFnEndTag
};

static LMEErr newWriteCB(LMEInt fd, LMEConstText buffer, LMEInt length, void *userData)
{
	if (fd < gFirstFreeId)
		return oldWriteCB(fd, buffer, length, userData);
	
	if (fd < gFirstFreeId || fd >= gFirstFreeId + maxc
									|| !c[fd - gFirstFreeId].inUse)
		return kLMEErrBadFd;
	if (KSend(&c[fd - gFirstFreeId].ref, buffer, length) != kKhepNoErr)
		return kLMEErrBadAccess;
	return kLMEErrOk;
}

static LMEErr newReadCB(LMEInt fd, LMEText buffer, LMEInt maxlength, LMEInt *length,
																		void *userData)
{
	int err;
	LMEInt i, j, k;
	
	if (fd < gFirstFreeId)
		return oldReadCB(fd, buffer, maxlength, length, userData);
	
	if (fd < gFirstFreeId || fd >= gFirstFreeId + maxc
									|| !c[fd - gFirstFreeId].inUse)
		return kLMEErrBadFd;
	for (i = 0; i < maxlength; )
	{
		err = KReceive(&c[fd - gFirstFreeId].ref, buffer + i, maxlength - i, length);
		if (err != kKhepNoErr)
			break;
		if (c[fd - gFirstFreeId].textMode)
		{
			for (j = k = 0; j < *length; j++)
			{
				if (buffer[i + j] == 13)
				{
					buffer[i + k++] = 10;
					c[fd - gFirstFreeId].wasCR = 1;
				}
				else
				{
					if (buffer[i + j] != 10 || !c[fd - gFirstFreeId].wasCR)
						buffer[i + k++] = buffer[i + j];
					c[fd - gFirstFreeId].wasCR = 0;
				}
			}
			*length = k;
		}
		i += *length;
	}
	switch (err)
	{
		case kKhepNoErr:
			*length = i;
			return kLMEErrOk;
		case kKhepComTimeoutErr:
			return kLMEErrTimeout;
		default:
			return kLMEErrBadAccess;
	}
}

LMEInt GetSQExtensionsVersion(void)
{
	return kCurrentVersion;
}

void InstallSQExtensions(LMEInt features,
								LMEXFunctionStruct **xfn,
								LMEWriteCB *writeCB, void **writeData,
								LMEReadCB *readCB, void **readData,
								LMEReadFunctionCB *readFnCB, void **readFnData,
								LMEInt firstFreeId)
	// entry point of the shared library
	// xfn should point to an array of extension functions whose last
	//  element is LMEXFnEndTag, or be 0
	// the callback should be set to new callbacks
	//  (called with old callbacks or 0 if none)
{
	LMEText msg;
	LMEInt msgLength;
	(void)readData, readFnCB, readFnData;
	
	if (features & 1)
	{
		*xfn = optXf;
			
		// display a message to stdout
		
		msg = "Extensions for K-Team devices\r"
				"(kserialname, kopen, ksend, kflush, kclose, opendevice, closedevice)\r"
				"and standard set of unsupported extensions:\r"
				"openshlib, callshlib, callshlib_old, closeshlib, launchurl\r"
				"Copyright 1997-2000, Yves Piguet.\r\r";
		for (msgLength = 0; msg[msgLength]; msgLength++)
			;
		if (*writeCB)
			(*writeCB)(1, msg, msgLength, *writeData);
		
		gFirstFreeId = firstFreeId;
		oldWriteCB = *writeCB;
		*writeCB = newWriteCB;
		oldReadCB = *readCB;
		*readCB = newReadCB;
		
		InitDevices();
	}
	else
	{
		// display a message to stdout
		
		msg = "No license for extensions for K-Team devices\r\r";
		for (msgLength = 0; msg[msgLength]; msgLength++)
			;
		if (*writeCB)
			(*writeCB)(1, msg, msgLength, *writeData);
	}
	
}

void ShutdownSQExtensions(void)
{
	KShutdown();
	ShutdownDevices();
}
